from math import atan, degrees
import pyqtgraph as pg
from PyQt5.QtCore import *
import pandas as pd
from cfatools.logreader.pump_calib import PumpCalibReader
import cfatools.processor.pump as pump_proc

from config import Config
from gui.uimainwindow import Ui_MainWindow


class PumpCalibUim:

    def __init__(self, main_ui: Ui_MainWindow, config: Config):
        self.main_ui = main_ui
        self.config = config

        self.pump_reader = self.set_current_pump_reader()
        self.main_ui.pump_combobox_pump.currentTextChanged.connect(self.change_pump)
        self.change_pump(self.main_ui.pump_combobox_pump.currentText())

        self.runs_df = pd.DataFrame()
        self.coef_df = pd.DataFrame()

    def set_current_pump_reader(self, pump_text: str = None) -> PumpCalibReader:
        if pump_text is None:
            pump_text = self.main_ui.pump_combobox_pump.currentText()

        if pump_text == "Common pump":
            self.pump_reader = PumpCalibReader(self.config.read("DATA_SOURCE", "absolute_root_dir"), "COMPUMP")
        elif pump_text == "Collector pump":
            self.pump_reader = PumpCalibReader(self.config.read("DATA_SOURCE", "absolute_root_dir"), "COLPUMP")
        elif pump_text == "Picarro pump":
            self.pump_reader = PumpCalibReader(self.config.read("DATA_SOURCE", "absolute_root_dir"), "PICPUMP")
        else:
            raise ValueError("Impossible to assign a Provider from text [" + pump_text + "]")

        return self.pump_reader

    def change_pump(self, pump_text: str = None):
        self.set_current_pump_reader(pump_text)
        self.runs_df = self.pump_reader.get_runs_df()
        self.coef_df = self.pump_reader.get_coef_df()
        self.__update_runs_plot__(self.runs_df, self.coef_df)

        turns_history_df = pump_proc.get_turns_history_df(
            rpm_history_df=self.pump_reader.get_state_changes_df("rpm"),
            turning_history_df=self.pump_reader.get_state_changes_df("turning")
        )
        evol_df = pump_proc.get_mlmin_to_rpm_evol_df(turns_history_df=turns_history_df, runs_df=self.runs_df)
        self.__update_evol_plot__(evol_df)

    ####################################################################
    # Plot

    def __update_runs_plot__(self, runs_df: pd.DataFrame, coef_df: pd.DataFrame) -> None:
        # Create plot item
        plot_item = pg.PlotItem(labels={"left": "ml/min", "bottom": "RPM"})
        self.main_ui.pump_graphicsview_runs.setCentralItem(plot_item)

        # Create legend
        legend = pg.LegendItem(offset=(57, 5))
        legend.setParentItem(plot_item.graphicsItem())

        # Create the color scale
        color_map = pg.ColorMap(pos=[runs_df["calib_id_int"].min(),
                                     runs_df["calib_id_int"].max()],
                                color=[[0, 0, 255], [0, 255, 0]])

        # Add linear regression line
        for row_id, row in coef_df.iterrows():
            reg_line = pg.InfiniteLine(pos=QPointF(0, row["intercept"]),
                                       angle=degrees(atan(row["slope"])),
                                       pen=pg.mkPen(color=color_map.mapToQColor(row["calib_id_int"])),
                                       name=row["calib_id_str"])
            plot_item.addItem(reg_line)

        # Add runs points
        for calib_id in runs_df["calib_id_int"].unique():
            run_df = runs_df[runs_df["calib_id_int"] == calib_id]
            run_points = pg.ScatterPlotItem(run_df["rpm"],
                                            run_df["flow"],
                                            size=6,
                                            brush=color_map.mapToQColor(calib_id),
                                            name=run_df["calib_id_str"].unique()[0])
            legend.addItem(run_points, run_df["calib_id_str"].unique()[0])
            plot_item.addItem(run_points)

    def __update_evol_plot__(self, evol_df: pd.DataFrame) -> None:
        # Create plot item
        plot_item = pg.PlotItem(labels={"left": "ml/min", "bottom": "Total rounds"})
        self.main_ui.pump_graphicsview_repro.setCentralItem(plot_item)

        # Create legend
        legend = pg.LegendItem(offset=(57, 5))
        legend.setParentItem(plot_item.graphicsItem())

        # Create the color scale
        color_map = pg.ColorMap(pos=[evol_df["rpm"].min(),
                                     evol_df["rpm"].max()],
                                color=[[255, 0, 255], [255, 0, 0]])

        for rpm in evol_df["rpm"].unique():
            rpm_df = evol_df[evol_df["rpm"] == rpm]
            rpm_points = pg.ScatterPlotItem(rpm_df["total_turns"],
                                            rpm_df["flow"],
                                            size=6,
                                            brush=color_map.mapToQColor(rpm),
                                            name=rpm)
            legend.addItem(rpm_points, rpm)
            plot_item.addItem(rpm_points)