import pyqtgraph as pg
import pyqtgraph.exporters
import datetime
import pandas as pd
import numpy as np
import re
import copy
from PyQt5.QtWidgets import QTableWidget, QFileDialog
from PyQt5.QtGui import QColor

datetime_format = "%Y-%m-%dT%H:%M:%S.%fZ"


def pd_time_to_epoch_ms(series: pd.Series) -> list:
    """Convert a Pandas' Series containing Timestamp data into 1D array containing seconds (as float) since epoch."""
    # Pandas' Timestamp data are in nanoseconds.
    return list(pd.to_datetime(series).astype(np.int64) / 1e9)


def datetime64_to_epoch_ms(datetimes_list: np.ndarray) -> list:
    return [float(instant) / 1e9 for instant in datetimes_list]


def convert_to_numeric(str_value: str):
    """Convert a string to float (if there is a "." in the string) or to integer, if possible."""
    if re.match("^-?\d+\.\d+$", str_value):
        return float(str_value)
    elif re.match("^-?\d+$", str_value):
        return int(str_value)
    else:
        return str_value


####################################################################
# Graphics

class TimeAxisItem(pg.AxisItem):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.setLabel(text='Time', units=None)
        self.enableAutoSIPrefix(False)

    def tickStrings(self, values, scale, spacing):
        try:
            tick_strings = [datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc).strftime("%H:%M:%S") for value in values]
        except ValueError:
            tick_strings = ["ERROR" for value in values]
        return tick_strings


def clear_legend_items(legend_item: pg.LegendItem) -> pg.LegendItem:
    """Remove all the elements of the input `legend_item`.

    Parameters
    ----------
    legend_item: pg.LegendItem
        Legend item to be emptied.

    Returns
    -------
    pg.LegendItem:
        Same as input, but empty.
    """
    for sample, label in legend_item.items:
        legend_item.removeItem(label.text)
    return legend_item


def export_plot(plot_item: pg.PlotItem, graphics_view: pg.PlotWidget) -> None:
    """Export the `plot_item` as a PNG file.

    This is a convenience function for exporting plots with pre-set parameters. The same thing can be achieved by
    right-clicking the plot and using build-in pyqtgraph's contextual menu. Except that the axis' color cannot be
    modified with the right-click button.

    Parameters
    ----------
    plot_item: pg.PlotItem
        Plot item to be exported as PNG file.
    graphics_view: pg.PlotWidget
        Widget where the `plot_item` is. Used to determine plot's dimension in the GUI.

    Returns
    -------
    None
        Creates a PNG file. A file dialog asks the user the output file name/path.
    """
    filename = QFileDialog.getSaveFileName(caption="Export plot as PNG", filter="*.png")

    exporter = pg.exporters.ImageExporter(plot_item)

    exporter.params.param('width').setValue(graphics_view.frameGeometry().width(),
                                            blockSignal=exporter.widthChanged)
    exporter.params.param('height').setValue(graphics_view.frameGeometry().height(),
                                             blockSignal=exporter.heightChanged)
    exporter.params.param('background').setValue(QColor("white"))

    # Change Axis' pen to have it black (more readable on white background than the original grey)
    gui_pen = plot_item.getAxis('left').pen()
    png_pen = pg.mkPen(color=(0, 0, 0))
    plot_item.getAxis('left').setPen(png_pen)
    plot_item.getAxis('bottom').setPen(png_pen)

    # Save to file
    # /!\ Note: there is a bug in pyqtgraph version 0.10.0, which lead to the "TypeError: 'float' object cannot be
    # interpreted as an integer" error.
    # Line 70 of file pyqtgraph.exporters.ImageExporter should be modified as follows:
    # bg = np.empty((int(self.params['width']), int(self.params['height']), 4), dtype=np.ubyte)
    exporter.export(filename[0])

    # Reset Axis' pen to previous pen
    plot_item.getAxis('left').setPen(gui_pen)
    plot_item.getAxis('bottom').setPen(gui_pen)


def highlight_row(table: QTableWidget, highlighted_row_id: int,
                  highlighted_row_color: QColor = QColor(173, 216, 230),
                  other_rows_color: QColor = QColor(255, 255, 255)) -> None:
    """Highlight a table's row by changing its background color.

    Parameters
    ----------
    table: PyQt5.QtWidgets.QTableWidget
        Table whose row should be highlighted.
    highlighted_row_id: int
        Index (first row is 0) of the row which should be highlighted.
    highlighted_row_color: PyQt5.QtGui.QColor
        Background color of the highlighted row. Default is light blue.
    other_rows_color: PyQt5.QtGui.QColor
        Background color of the other rows of the table. Default is white.

    Returns
    -------
        Nothing, modify the input ``table``
    """
    for row_id in range(table.rowCount()):
        if row_id == highlighted_row_id:
            row_color = highlighted_row_color
        else:
            row_color = other_rows_color
        for column_id in range(table.columnCount()):
            table.item(row_id, column_id).setBackground(row_color)